/* main.c - Boot sector program. */
/* Copyright (C) 2017 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/********************************************************************/
/*                                                                  */
/*			   Project Includes                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   System Includes                          */
/*                                                                  */
/********************************************************************/
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

/********************************************************************/
/*                                                                  */
/*			  Macro Definitions                         */
/*                                                                  */
/********************************************************************/
#define SECTOR_SIZE 0x200
/********************************************************************/
/*                                                                  */
/*			      Data Types                            */
/*                                                                  */
/********************************************************************/
/*  
 * Type definition and structure for the address packet used for LBA
 * disk addressing.
 */
typedef struct address_packet_struct address_packet_T;
struct address_packet_struct
{
  uint8_t size;			/* Size of this address packet. */
  uint8_t mbz;			/* Must be zero. */
  uint16_t sectors;		/* Number of sectors to transfer. */
  uint32_t buffer;		/* Transfer buffer - see note #1. */
  uint32_t lba_lower;		/* Lower 32 bits of starting LBA. */
  uint32_t lba_upper;		/* Upper 32 bits of starting LBA. */
} __attribute__((__packed__));
/* Notes: --------------------------------------------------------- */
/*   
 *   1.  The segment is in the upper 16 bits of `buffer', and the
 *       offset is in the lower 16 bits of `buffer'.
 */
typedef struct packed_int_struct packed_int_T;
struct packed_int_struct
{
  uint32_t lba_lower;
  uint32_t lba_upper;
} __attribute__((__packed__));
/*
 * Type definition and union for LBA disk address.
 */
typedef union disk_address_union disk_address_T;
union disk_address_union
{
  uint64_t l;
  packed_int_T dwords;
};

/*
 * Type definition and structure for segment/offset.
 */
typedef struct segment_struct segment_T;
struct segment_struct
{
  uint16_t offset;
  uint16_t segment;
} __attribute__ ((__packed__));

/*
 * Type definition and union for calculating segment/offset from a
 * pointer;
 */
typedef union address_calculator_union address_calculator_T;
union address_calculator_union
{
  void *ptr;
  uint32_t i;
  segment_T segoff;
};
/*
 * Type definition for loader "function"  
 */
typedef void (*loader_T) (void);
/*
 * Type definition and structure for load image table (LIT).
 */
typedef struct load_table_struct load_table_T;
struct load_table_struct
{
  loader_T entry;
  void *edata;
} __attribute__((packed));

/********************************************************************/
/*                                                                  */
/*			   Global Variables                         */
/*                                                                  */
/********************************************************************/
/* Implemented in the linker script `linker.ld', this is the loader
 * information table.
 */
extern load_table_T lit;

/********************************************************************/
/*                                                                  */
/*			   Module Variables                         */
/*                                                                  */
/********************************************************************/
/*
 * This is the disk address packet (DAP).
 */
__attribute__((__aligned__(4) )) static address_packet_T dap;

/********************************************************************/
/*                                                                  */
/*			 Interface Functions                        */
/*                                                                  */
/********************************************************************/
/*
 * The main driver function for the boot sector program. 
 */
void cmain (void);

/********************************************************************/
/*                                                                  */
/*			   Module Functions                         */
/*                                                                  */
/********************************************************************/
/* Text output functions........................................... */
/*
 * Prints a single character CH to the console. 
 */
static void putchar (char ch);

/*
 * Prints string S to the console.
 */
static void puts (char *s);

/*
 * Prints the error message specified in MESSAGE and halts the
 * processor.
 */
__attribute__((__noreturn__)) static void error (char *message);

/*  Loader functions............................................... */
/*
 * Tests if LBA extensions are available.  If LBA extensions are
 * supported, returns true, otherwise it returns false.
 */
static bool test_lba (void);

/*
 * Reads drive.  If successful, returns true, otherwise it returns
 * false.
 */
static bool read_drive (uint64_t lba, uint16_t sectors, void *buffer);

/*
 * Set up DAP. 
 */
static void setup_dap (uint64_t lba, uint16_t sectors, void *buffer);
/********************************************************************/
/*                                                                  */
/*			   Inline Functions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			    Implementation                          */
/*                                                                  */
/********************************************************************/
void 
cmain (void)
{
  uint16_t sectors;

  if (test_lba () == false)
    error ("LBA not supported");
  if (read_drive (1, 1, &lit) == false)
    error ("Unable to read drive");
  sectors = ((uint32_t) lit.edata - (uint32_t) &lit) / 0x200 - 1;
  if (sectors > 0)
    {
      if (read_drive (2, sectors, (void *) &lit + SECTOR_SIZE) == false)
  	error ("Unable to read drive");
    }
  lit.entry ();
}
static void 
putchar (char ch)
{
#define COLOR 0x7f
  int16_t buffer = 0x0e00;
  
  buffer += ch;
  asm ("int $(0x10)" 
       : 
       : "a" (buffer), "b" ((uint16_t) COLOR)
       :
       );
}
static void
puts (char *s)
{
  char *ptr;
  for (ptr = s; *ptr != '\0'; ++ptr)
    putchar (*ptr);
}
__attribute__((noreturn)) static void 
error (char *message)
{
  puts (message);
  puts (": processor halted");
  for (;;)
    asm ("hlt");
}
static bool
test_lba (void)
{
  bool ret = true;
  
  asm goto ("int $(0x10); jc %l3"
	    : 
	    : "a" ((uint16_t) 0x4100), "b" ((uint16_t) 0x55aa), "d" ((uint8_t) 0x80)
	    :
	    : lba_unsupported
	    );
 quit:
  return ret;
 lba_unsupported:
  asm ("clc");
  ret = false;
  goto quit;
}
static bool
read_drive (uint64_t lba, uint16_t sectors, void *buffer)
{
  bool ret = true;
  
  setup_dap (lba, sectors, buffer);
  asm goto ("int $(0x13); jc %l3"
	    :
	    : "a" ((uint16_t) 0x4200), "d" ((uint8_t) 0x80), "S" (&dap)
	    :
	    : read_error
	    );
 quit:
  return ret;
 read_error:
  ret = false;
  goto quit;
}
static void
setup_dap (uint64_t lba, uint16_t sectors, void *buffer)
{
  disk_address_T addr;
  address_calculator_T calculator;

  /*
   * Setup simple fields in DAP. 
   */
  dap.size = sizeof (address_packet_T);
  dap.mbz = 0;
  dap.sectors = sectors;

  /*
   * Setup buffer.
   */
  calculator.ptr = buffer;
  calculator.segoff.segment <<= 12;
  dap.buffer = calculator.i;

  /*
   * Setup LBA address. 
   */
  addr.l = lba;
  dap.lba_lower = addr.dwords.lba_lower;
  dap.lba_upper = addr.dwords.lba_upper;
}
