/* console.c - Implementation of console driver. */
/* Copyright (C) 2017 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif

/********************************************************************/
/*                                                                  */
/*			   Project Includes                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   System Includes                          */
/*                                                                  */
/********************************************************************/
#include <stdarg.h>
#include <stdint.h>
/********************************************************************/
/*                                                                  */
/*			  Macro Definitions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			      Data Types                            */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   Global Variables                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			   Module Variables                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			 Interface Functions                        */
/*                                                                  */
/********************************************************************/
void printf (const char *format, ...);
void vprintf (const char *format, va_list arg);
__attribute__((__noreturn__)) void error (const char *format, ...);
__attribute__((__noreturn__)) void verror (const char *format, va_list arg);
char getchar (void);
/********************************************************************/
/*                                                                  */
/*			   Module Functions                         */
/*                                                                  */
/********************************************************************/
/*
 * Internal implementation of printf function.  
 */
static void printf_internal (const char *format, ...);
/* 
 * Internal implementation of vprintf function. 
 */
static void vprintf_internal (const char *format, va_list arg);
/*
 * Converts string to integer.
 */
static void itoa (char *buf, int32_t base, int32_t d);
/*
 * Prints a single character `ch' to the console. 
 */
static void putchar (char ch);
/*
 * Reads a single character from the console (keyboard).
 */
static char getchar_internal (void);
/********************************************************************/
/*                                                                  */
/*			   Inline Functions                         */
/*                                                                  */
/********************************************************************/
/********************************************************************/
/*                                                                  */
/*			    Implementation                          */
/*                                                                  */
/********************************************************************/
void
printf (const char *format, ...)
{
  va_list arg;
  
  va_start (arg, format);
  vprintf_internal (format, arg);
}
void
vprintf (const char *format, va_list arg)
{
  vprintf_internal (format, arg);
}
__attribute__((__noreturn__)) void
error (const char *format, ...)
{
  va_list arg;

  va_start (arg, format);
  vprintf_internal (format, arg);
  printf_internal (": processor halted.");
  for (;;)
    asm ("hlt");
}
__attribute__((__noreturn__)) void
verror (const char *format, va_list arg) 
{
  vprintf_internal (format, arg);
  printf_internal (": processor halted.");
  for (;;)
    asm ("hlt");
}
char
getchar (void)
{
  char ret;

  ret = getchar_internal ();
  return ret;
}
static void
printf_internal (const char *format, ...)
{
  va_list arg;
  
  va_start (arg, format);
  vprintf_internal (format, arg);
}
static void
vprintf_internal (const char *format, va_list arg)
{
  char c;
  static char buf[20];
  
  while ((c = *format++) != 0)
    {
      if (c != '%')
	putchar (c);
      else
	{
	  char *p;
     
	  c = *format++;
	  switch (c)
	    {
	    case 'd':
	    case 'u':
	    case 'x':
	      itoa (buf, c, va_arg (arg, int32_t));
	      p = buf;
	      goto string;
	      break;
	    case 's':
	      p = va_arg (arg, char *);
	      if (! p)
		p = "(null)";
	    string:
	      while (*p)
		putchar (*p++);
	      break;
	    default:
	      putchar (va_arg (arg, int32_t));
	      break;
	    }
	}
    }
}
static void
itoa (char *buf, int32_t base, int32_t d)
{
  char *p = buf;
  char *p1, *p2;
  uint32_t ud = d;
  int32_t divisor = 10;
     
  /* If %d is specified and D is minus, put `-' in the head. */
  if (base == 'd' && d < 0)
    {
      *p++ = '-';
      buf++;
      ud = -d;
    }
  else if (base == 'x')
    divisor = 16;
     
  /* Divide UD by DIVISOR until UD == 0. */
  do
    {
      int32_t remainder = ud % divisor;
     
      *p++ = (remainder < 10) ? remainder + '0' : remainder + 'a' - 10;
    }
  while (ud /= divisor);
     
  /* Terminate BUF. */
  *p = 0;
     
  /* Reverse BUF. */
  p1 = buf;
  p2 = p - 1;
  while (p1 < p2)
    {
      char tmp = *p1;
      *p1 = *p2;
      *p2 = tmp;
      p1++;
      p2--;
    }
}
static void
putchar (char ch)
{
#define COLOR 0x7f
  int16_t buffer = 0x0e00;
  
  buffer += ch;
  asm ("int $(0x10)"
       :
       : "a" (buffer), "b" ((uint16_t) COLOR)
       :
       );
}
static char
getchar_internal (void)
{
  char ret;

  asm ("int $0x16"
       : "=a" (ret)
       : "a" ((uint16_t) 0)
       :
       );
  putchar (ret);
  return ret;
}
